# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import MonitorDaemon, schema
from cmf.monitor import constants
from cmf.monitor.datanode import DataNodeMonitor
import cmf.monitor.datanode.json_parser
import cmf.monitor.datanode.json_parser_jmx
from cmf.monitor.tests import make_config_file

from nose.tools import assert_true, assert_equals, assert_false

class MockDataNodeMonitor(DataNodeMonitor):
  def __init__(self, daemon, registry, string):
    DataNodeMonitor.__init__(self, daemon, registry)
    self.set_string(string)
    self.set_throw_error(None)

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url):
    if self._error:
      raise self._error
    return StringIO(self._string)

# CM 3.7 config file format: no service version.
def _make_cm37_format_config_file(version, dfs_datanode_http_port):
  properties = """[datanode]
role_name = hdfs-DN
service_name = my_hdfs
host = localhost
dfs_datanode_http_port = %s
dfs_data_dirs = /path/to/dir,/another/data/dir
log_dir = /path/to/logs
collect_interval = 1.0
"""
  return  make_config_file(properties % (dfs_datanode_http_port,))

# CM 4.0 config file: we added the service version.
def _make_cm40_format_config_file(version, dfs_datanode_http_port):
  properties = """[datanode]
role_name = hdfs-DN
service_name = my_hdfs
service_version = %s
host = localhost
dfs_datanode_http_port = %s
dfs_data_dirs = /path/to/dir,/another/data/dir
log_dir = /path/to/logs
collect_interval = 1.0
"""
  return  make_config_file(properties % (version, dfs_datanode_http_port,))

def test_dn_lifecycle_cdh3():
  do_dn_lifecycle(3, _make_cm37_format_config_file,
                  cmf.monitor.datanode.json_parser.SAMPLE)
  do_dn_lifecycle(3, _make_cm40_format_config_file,
                  cmf.monitor.datanode.json_parser.SAMPLE)

def test_dn_lifecycle_cdh4():
  do_dn_lifecycle(4, _make_cm40_format_config_file,
                  cmf.monitor.datanode.json_parser_jmx.SAMPLE)

def do_dn_lifecycle(version, make_config_file_function, json_sample):
  daemon = MonitorDaemon("host", "host", agent=None)
  reg = daemon.get_registry_for_testing()
  monitor = MockDataNodeMonitor(daemon, reg, json_sample)
  monitor.refresh_pid(os.getpid(), True)

  # Make two different configs
  conf_dir = make_config_file_function(version, 50075)
  conf_dir2 = make_config_file_function(version, 50076)

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    assert_true('50075' in monitor.metrics_url, "Config installed")

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()
    assert_true('50076' in monitor.metrics_url, "Using the new config")
    assert_equals(['/path/to/dir', '/another/data/dir'],
                  monitor.dfs_data_dirs)

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('hdfs-DN', role_update['rolename'])
    assert_equals('my_hdfs', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_equals(8, metric_dict[schema.DN_NUM_REMOTE_READS])
    assert_equals(4, metric_dict[schema.DN_BYTES_WRITTEN])
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0) 
    assert_equals(0, metric_dict[schema.DN_NUM_VOLUME_FAILURES])
    assert_equals(888.9375, metric_dict[schema.JVM_MAX_MEMORY_MB])
    assert_equals(1, metric_dict[schema.JVM_BLOCKED_THREADS])
    assert_equals(2, metric_dict[schema.JVM_NEW_THREADS])
    assert_equals(3, metric_dict[schema.JVM_RUNNABLE_THREADS])
    assert_equals(4, metric_dict[schema.JVM_TERMINATED_THREADS])
    assert_equals(5, metric_dict[schema.JVM_TIMED_WAITING_THREADS])
    assert_equals(6, metric_dict[schema.JVM_WAITING_THREADS])
    assert_equals(21, metric_dict[schema.JVM_TOTAL_THREADS])
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test fs update
    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 3, "We configured 3 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_equals('/path/to/dir', dir_metrics[1]['path'])
    assert_equals('/another/data/dir', dir_metrics[2]['path'])

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('hdfs-DN', role_update['rolename'])
    assert_equals('my_hdfs', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 3, "We configured 3 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_equals('/path/to/dir', dir_metrics[1]['path'])
    assert_equals('/another/data/dir', dir_metrics[2]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)
